/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <glib.h>
#include "nntpgrab_utils.h"

struct _parse_data {
    NNTPGrabNZB *nzb;
    NNTPGrabNZBFile *current_file;
    NNTPGrabPart *current_segment;
};

static void
start_element_handler  (GMarkupParseContext *context,
                        const gchar         *element_name,
                        const gchar        **attribute_names,
                        const gchar        **attribute_values,
                        gpointer             user_data,
                        GError             **error)
{
    struct _parse_data *data = (struct _parse_data*) user_data;
    int i;

    if (!strcmp(element_name, "file")) {
        NNTPGrabNZBFile *file = g_slice_new0(NNTPGrabNZBFile);

        g_return_if_fail(data->current_file == NULL);

        data->nzb->files = ng_list_append(data->nzb->files, file);
        data->current_file = file;

        i = 0;
        while (attribute_names[i] != NULL) {
            if (!strcmp(attribute_names[i], "subject")) {
                strncpy((char *) file->subject, attribute_values[i], (size_t) sizeof(file->subject));
            } else if (!strcmp(attribute_names[i], "poster")) {
                strncpy((char *) file->poster, attribute_values[i], (size_t) sizeof(file->poster));
            } else if (!strcmp(attribute_names[i], "date")) {
                file->stamp = atoi(attribute_values[i]);
            }

            i++;
        }
    } else if (!strcmp(element_name, "segment")) {
        NNTPGrabPart *segment = g_slice_new0(NNTPGrabPart);

        g_return_if_fail(data->current_file != NULL);

        data->current_file->segments = ng_list_append(data->current_file->segments, segment);
        data->current_segment = segment;

        i = 0;
        while (attribute_names[i] != NULL) {
            if (!strcmp(attribute_names[i], "bytes")) {
                segment->size = atol(attribute_values[i]);
                data->current_file->filesize += segment->size;
            } else if (!strcmp(attribute_names[i], "number")) {
                segment->part_num = atoi(attribute_values[i]);
            }

            i++;
        }
    }
}

static void
end_element_handler    (GMarkupParseContext *context,
                        const gchar         *element_name,
                        gpointer             user_data,
                        GError             **error)
{
    struct _parse_data *data = (struct _parse_data*) user_data;

    if (!strcmp(element_name, "file")) {
        data->current_file = NULL;
    } else if (!strcmp(element_name, "segment")) {
        data->current_segment = NULL;
    }
}

static void
text_handler           (GMarkupParseContext *context,
                        const gchar         *text,
                        gsize                text_len,
                        gpointer             user_data,
                        GError             **error)
{
    struct _parse_data *data = (struct _parse_data*) user_data;
    const char *element_name = g_markup_parse_context_get_element(context);

    if (!strcmp(element_name, "segment")) {
        char *message_id;

        g_return_if_fail(data->current_segment != NULL);

        /* NZB files contain a message-id without the brackets. Add them manually */
        message_id = g_strdup_printf("<%s>", text);
        strncpy((char*) data->current_segment->message_id, message_id, (size_t) sizeof(data->current_segment->message_id) - 1);
        g_free(message_id);
    } else if (!strcmp(element_name, "group")) {
        g_return_if_fail(data->current_file != NULL);

        data->current_file->groups = ng_list_append(data->current_file->groups, g_strdup(text));
    }
}

static const GMarkupParser parser = {
  start_element_handler,
  end_element_handler,
  text_handler,
  NULL,
  NULL
};

NNTPGrabNZB *
nntpgrab_utils_parse_nzb_file(const char *contents, char **errmsg)
{
    GMarkupParseContext *context;
    GError              *error = NULL;
    struct _parse_data   parse_data;

    parse_data.nzb = g_slice_new0(NNTPGrabNZB);
    parse_data.current_file = NULL;
    parse_data.current_segment = NULL;

    context = g_markup_parse_context_new (&parser, 0, &parse_data, NULL);

    if (!g_markup_parse_context_parse (context, contents, -1, &error) ||
        !g_markup_parse_context_end_parse (context, &error)) {

        if (g_error_matches(error, G_MARKUP_ERROR, G_MARKUP_ERROR_BAD_UTF8)) {
            /* Encoding error. Try converting the file from iso-8859-1 to utf-8 first */
            char *contents2 = g_convert(contents, -1, "utf-8", "iso-8859-1", NULL, NULL, NULL);

            g_error_free(error);
            error = NULL;

            nntpgrab_utils_nzb_file_free(parse_data.nzb);
            parse_data.nzb = g_slice_new0(NNTPGrabNZB);
            parse_data.current_file = NULL;
            parse_data.current_segment = NULL;

            g_markup_parse_context_free (context);
            context = g_markup_parse_context_new (&parser, 0, &parse_data, NULL);

            g_markup_parse_context_parse (context, contents2, -1, &error) && g_markup_parse_context_end_parse (context, &error);

            g_free(contents2);
        }

        if (error) {
            if (errmsg) {
                *errmsg = g_strdup(error->message);
            }
            g_error_free(error);
            g_markup_parse_context_free (context);
            nntpgrab_utils_nzb_file_free(parse_data.nzb);

            return NULL;
        }
    }

    g_markup_parse_context_free (context);

    return parse_data.nzb;
}

void
nntpgrab_utils_nzb_file_free(NNTPGrabNZB *nzbfile)
{
    NGList *list;

    list = nzbfile->files;
    while (list) {
        NNTPGrabNZBFile *file = (NNTPGrabNZBFile *) list->data;
        NGList *list2;

        list2 = file->groups;
        while (list2) {
            g_free(list2->data);
            list2 = ng_list_next(list2);
        }

        ng_list_free(file->groups);

        list2 = file->segments;
        while (list2) {
            g_slice_free(NNTPGrabPart, list2->data);
            list2 = ng_list_next(list2);
        }

        ng_list_free(file->segments);

        g_slice_free(NNTPGrabNZBFile, file);

        list = ng_list_next(list);
    }

    g_slice_free(NNTPGrabNZB, nzbfile);
}
