/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "gui.h"
#include "config_gui.h"
#include "nntpgrab_automation.h"
#ifdef HAVE_LIBNOTIFY
#include <libnotify/notify.h>
#endif

static NNTPGrabAutoImport *auto_import = NULL;

static void
on_nzb_imported(NNTPGrabAutoImport *auto_import, const char *filename, const char *collection_name, gpointer data)
{
#ifdef HAVE_LIBNOTIFY
    NotifyNotification *notification;
    char *msg;

    if (auto_import) {
        /* The frontend has imported an NZB file */
        msg = g_strdup_printf(_("The file '%s'\nwas successfully imported with the collection name '%s'"), filename, collection_name);
    } else {
        /* The backend (which could be a different host) has imported an NZB file */
        if (nntpgrab_glue_get_is_standalone(glue)) {
            msg = g_strdup_printf(_("The file '%s'\nwas successfully imported with the collection name '%s'"), filename, collection_name);
        } else {
            msg = g_strdup_printf(_("The file '%s'\non the NNTPGrab Server was successfully\nimported with the collection name '%s'"), filename, collection_name);
        }
    }
    notification = notify_notification_new(_("NZB File Imported"), msg, NULL, NULL);
    g_free(msg);
    notify_notification_show(notification, NULL);
#endif
}

static void
on_plugin_event(NNTPGrabGlue *glue, const char *plugin_name, const char *event_name, const char **params, gpointer data)
{
    if (!strcmp(plugin_name, "Auto-import") && !strcmp(event_name, "nzb_imported")) {
        on_nzb_imported(NULL, params[0], params[1], data);
    }
}

void
auto_import_import_nzb_file(NNTPGrabGlue *glue, NNTPGrabNZB *nzbfile, const char *collection_name, char **warnings)
{
    NGList *list;
    char *errmsg = NULL;

    list = nzbfile->files;

    if (warnings) {
        *warnings = NULL;
    }

    while (list) {
        NNTPGrabNZBFile *file = list->data;

        if (!nntpgrab_schedular_add_task_to_queue(glue, collection_name, file->subject, file->poster, file->stamp, file->filesize, file->groups, file->segments, &errmsg)) {
            if (warnings) {
                if (*warnings) {
                    char *tmp = g_strdup_printf("%s\n%s", *warnings, errmsg);
                    g_free(*warnings);
                    *warnings = tmp;
                } else {
                    *warnings = g_strdup_printf(_("File could not be added to the download queue:\n%s"), errmsg);
                }
                g_free(errmsg);
            }
        }

        list = ng_list_next(list);
    }
}

void
enable_auto_import(void)
{
#if GLIB_CHECK_VERSION(2,16,0)
    char *errmsg = NULL;
    ConfigGUIOpts opts = config_gui_get_opts();

    if (!nntpgrab_automation_enable_auto_import(auto_import, opts.auto_import_directory, &errmsg)) {
        GtkWidget *windowMain = nntpgrab_gui_base_get_widget("windowMain");

        nntpgrab_gui_base_dialog_show(windowMain, errmsg, GTK_MESSAGE_INFO, GTK_BUTTONS_OK);

        g_free(errmsg);
    }
#endif
}

void
disable_auto_import(void)
{
#if GLIB_CHECK_VERSION(2,16,0)
    nntpgrab_automation_disable_auto_import(auto_import);
#endif
}

void
auto_import_initialize(void)
{
#if GLIB_CHECK_VERSION(2,16,0)
    ConfigGUIOpts opts = config_gui_get_opts();

    auto_import = nntpgrab_automation_auto_import_new((ConfigGetOptsFunc) nntpgrab_config_get_opts, (ImportNZBFileFunc) auto_import_import_nzb_file, glue);
    ng_signal_connect(auto_import, "nzb_imported", NG_CALLBACK(on_nzb_imported), NULL);

    ng_signal_connect(glue, "plugin_event", NG_CALLBACK(on_plugin_event), NULL);

#ifndef __APPLE__
    if (opts.enable_auto_import) {
        enable_auto_import();
    }
#endif
#endif
}

