/*
    Copyright (C) 2005-2009  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include "gui.h"
#include "config_gui.h"

static ConfigGUIOpts opts;
static GHashTable *active_connections = NULL;

static GtkTreeRowReference *
get_active_connection_row_ref(const char *servername, int conn_id)
{
    GtkTreeRowReference *ref;

    ref = g_hash_table_lookup(active_connections, GINT_TO_POINTER(conn_id));
    if (!ref) {
        GtkWidget *treeConnections;
        GtkTreeModel *store;
        GtkTreeIter iter;
        GtkTreePath *path;

        treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
        store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

        // No element found, add a new one
        gtk_list_store_append(GTK_LIST_STORE(store), &iter);
        gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, conn_id, 1, servername, 2, _("Connecting"), -1);

        path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);

        ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(store), path);
        g_hash_table_insert(active_connections, GINT_TO_POINTER(conn_id), ref);

        gtk_tree_path_free(path);
    }

    return ref;
}

static gboolean
get_connection_iter(GtkTreeModel *store, int conn_id, int line, GtkTreeIter *iter)
{
    GtkTreePath *path;

    /* Find the corresponding row in the tree */
    path = gtk_tree_path_new_from_indices(conn_id - 1, -1);
    if (!gtk_tree_model_get_iter(GTK_TREE_MODEL(store), iter, path)) {
        /* Shouldn't happen */
        gtk_tree_path_free(path);
        //g_warning(__FILE__ ":%i Unable to find connection with id %i in treeConnections", line, conn_id);
        return FALSE;
    }
    gtk_tree_path_free(path);

    return TRUE;
}

static void
connection_connecting (NNTPGrabGlue *glue, const char *servername, int conn_id, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    if (opts.hide_inactive_connections) {
        GtkTreePath *path;

        // Is this connection id already known?
        if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter)) {
            do {
                int id;

                gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, 0, &id, -1);

                if (id == conn_id) {
                    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 1, servername, 2, _("Connecting"), 4, "", -1);
                    return;
                }
            } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter));
        }

        // No element found, add a new one
        gtk_list_store_append(GTK_LIST_STORE(store), &iter);

        path = gtk_tree_model_get_path(GTK_TREE_MODEL(store), &iter);

        g_hash_table_insert(active_connections, GINT_TO_POINTER(conn_id), gtk_tree_row_reference_new(GTK_TREE_MODEL(store), path));

        gtk_tree_path_free(path);
    } else {
        if (!get_connection_iter(store, conn_id, __LINE__, &iter)) {
            return;
        }
    }

    /* Update the row */
    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, conn_id, 1, servername, 2, _("Connecting"), -1);
}

static void
connection_connected (NNTPGrabGlue *glue, const char *servername, int conn_id, const char *welcome_msg, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    if (opts.hide_inactive_connections) {
        GtkTreePath *path;
        GtkTreeRowReference *ref;

        ref = get_active_connection_row_ref(servername, conn_id);

        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(store, &iter, path);
        gtk_tree_path_free(path);
    } else {
        if (!get_connection_iter(store, conn_id, __LINE__, &iter)) {
            return;
        }
    }

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Connected"), 4, welcome_msg, -1);
}

static gboolean
update_disconnected_connection(gpointer data)
{
    GtkTreeRowReference *ref = (GtkTreeRowReference *) data;

    if (gtk_tree_row_reference_valid(ref)) {
        ConfigGUIOpts opts = config_gui_get_opts();
        GtkTreeModel *model = gtk_tree_row_reference_get_model(ref);
        GtkTreePath *path;
        GtkTreeIter iter;
        char *label = NULL;
        int conn_id;

        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(model, &iter, path);
        gtk_tree_path_free(path);

        gtk_tree_model_get(model, &iter, 0, &conn_id, 2, &label, -1);

        if (!strcmp(label, _("Disconnected"))) {
            if (opts.hide_inactive_connections) {
                // Connection is still inactive, remove it
                g_hash_table_remove(active_connections, GINT_TO_POINTER(conn_id));
            } else {
                gtk_list_store_set(GTK_LIST_STORE(model), &iter, 2, _("Unused"), 3, 0, 4, "", -1);
            }
        }

        g_free(label);
    }

    gtk_tree_row_reference_free(ref);

    return FALSE;
}

static void
connection_disconnect (NNTPGrabGlue *glue, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *disconnect_msg, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    GtkTreePath *path;
    GtkTreeRowReference *ref;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    if (opts.hide_inactive_connections) {
        ref = get_active_connection_row_ref(servername, conn_id);

        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(store, &iter, path);
        gtk_tree_path_free(path);
    } else {
        if (!get_connection_iter(store, conn_id, __LINE__, &iter)) {
            return;
        }
    }

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Disconnected"), 4, disconnect_msg, -1);

    // Add a timer so this record will automatically be removed after 10 seconds of inactivity
    path = gtk_tree_model_get_path(store, &iter);
    ref = gtk_tree_row_reference_new(store, path);
    gtk_tree_path_free(path);

#if GLIB_CHECK_VERSION(2,14,0)
    g_timeout_add_seconds(10, update_disconnected_connection, ref);
#else
    g_timeout_add(1000 * 10, update_disconnected_connection, ref);
#endif
}

static void
part_download_start (NNTPGrabGlue *glue, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    if (opts.hide_inactive_connections) {
        GtkTreePath *path;
        GtkTreeRowReference *ref;

        ref = get_active_connection_row_ref(servername, conn_id);
        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(store, &iter, path);
        gtk_tree_path_free(path);
    } else {
        if (!get_connection_iter(store, conn_id, __LINE__, &iter)) {
            return;
        }
    }

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Downloading"), 4, subject, -1);
}

static void
part_progress_update (NNTPGrabGlue *glue, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int bytes_downloaded, int bytes_total, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    double progress;
    int progress_int;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    if (opts.hide_inactive_connections) {
        GtkTreePath *path;
        GtkTreeRowReference *ref;

        ref = get_active_connection_row_ref(servername, conn_id);
        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(store, &iter, path);
        gtk_tree_path_free(path);
    } else {
        if (!get_connection_iter(store, conn_id, __LINE__, &iter)) {
            return;
        }
    }

    progress = bytes_downloaded;
    progress /= bytes_total;
    progress *= 100;
    progress_int = (int) progress;

    if (progress_int > 100) {
        g_print(__FILE__ ":%i progress overflow!\n", __LINE__);
        progress_int = 100;
    }

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Downloading"), 3, progress_int, -1);
}

static void
part_download_done (NNTPGrabGlue *glue, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    if (opts.hide_inactive_connections) {
        GtkTreePath *path;
        GtkTreeRowReference *ref;

        ref = get_active_connection_row_ref(servername, conn_id);
        path = gtk_tree_row_reference_get_path(ref);
        gtk_tree_model_get_iter(store, &iter, path);
        gtk_tree_path_free(path);
    }
    if (!get_connection_iter(store, conn_id, __LINE__, &iter)) {
        return;
    }

    gtk_list_store_set(GTK_LIST_STORE(store), &iter, 2, _("Idle"), 3, 0, 4, "", -1);
}

static void
config_changed(NNTPGrabGlue *glue, gpointer data)
{
    GtkWidget *treeConnections;
    GtkTreeModel *store;
    GtkTreeIter iter;
    NGList *servers;
    NGList *list;
    int conn_id = 0;

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeConnections));

    gtk_list_store_clear(GTK_LIST_STORE(store));

    if (opts.hide_inactive_connections) {
        /* Nothing to do */
        return;
    }

    servers = nntpgrab_config_get_avail_servers(glue);
    list = servers;
    while (list) {
        ConfigServer server;
        const char *servername = list->data;

        if (!nntpgrab_config_get_server_info(glue, servername, &server)) {
            /* Huh? shouldn't happen */
            g_warning(__FILE__ ":%i server information for server '%s' couldn't be found", __LINE__, servername);
            list = ng_list_next(list);
            continue;
        }

        if (server.enabled) {
            int i;

            for (i = 0; i < server.max_threads; i++) {
                conn_id++;
                gtk_list_store_append(GTK_LIST_STORE(store), &iter);
                gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, conn_id, 1, server.servername, 2, _("Unused"), -1);
            }
        }

        list = ng_list_next(list);
    }

    nntpgrab_config_free_avail_servers(glue, servers);
}

static void
connections_destroy_func(gpointer data)
{
    GtkTreeRowReference *ref = data;
    GtkTreePath *path;
    GtkTreeIter iter;
    GtkTreeModel *model;

    model = gtk_tree_row_reference_get_model(ref);
    path = gtk_tree_row_reference_get_path(ref);
    gtk_tree_model_get_iter(model, &iter, path);
    gtk_tree_path_free(path);

    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);

    gtk_tree_row_reference_free(ref);
}

void
connections_reread_config(void)
{
    opts = config_gui_get_opts();

    if (opts.hide_inactive_connections) {
        g_return_if_fail(active_connections == NULL);

        active_connections = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, connections_destroy_func);
    } else {
        if (active_connections) {
            g_hash_table_destroy(active_connections);
            active_connections = NULL;
        }
    }

    config_changed(glue, NULL);
}

void
connections_initialize(void)
{
    GtkWidget *treeConnections;
    GtkListStore *store;
    GtkCellRenderer *cellRenderer;

    opts = config_gui_get_opts();

    treeConnections = nntpgrab_gui_base_get_widget("treeConnections");

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Servername"), cellRenderer, "text", 1, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Status"), cellRenderer, "text", 2, NULL);

    cellRenderer = gtk_cell_renderer_progress_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Part progress"), cellRenderer, "value", 3, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeConnections), -1, _("Filename"), cellRenderer, "text", 4, NULL);

    store = gtk_list_store_new(5, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);
    gtk_tree_view_set_model(GTK_TREE_VIEW(treeConnections), GTK_TREE_MODEL(store));

    ng_signal_connect(glue, "config_changed", NG_CALLBACK(config_changed), NULL);
    ng_signal_connect(glue, "connection_connecting", NG_CALLBACK(connection_connecting), NULL);
    ng_signal_connect(glue, "connection_connected", NG_CALLBACK(connection_connected), NULL);
    ng_signal_connect(glue, "connection_disconnect", NG_CALLBACK(connection_disconnect), NULL);
    ng_signal_connect(glue, "part_download_start", NG_CALLBACK(part_download_start), NULL);
    ng_signal_connect(glue, "part_progress_update", NG_CALLBACK(part_progress_update), NULL);
    ng_signal_connect(glue, "part_done", NG_CALLBACK(part_download_done), NULL);
    ng_signal_connect(glue, "part_failed", NG_CALLBACK(part_download_done), NULL);

    connections_reread_config();
}
