/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include "gui.h"

#ifdef HAVE_DBUS
#include <dbus/dbus-glib.h>

typedef struct NNTPGrabDBusObject NNTPGrabDBusObject;

static NNTPGrabDBusObject *obj = NULL;
static DBusGProxy *bus_proxy = NULL;
static DBusGConnection *bus = NULL;

typedef struct NNTPGrabDBusObjectClass NNTPGrabDBusObjectClass;

GType nntpgrab_dbus_object_get_type (void);

struct NNTPGrabDBusObject
{
    GObject parent;
};

struct NNTPGrabDBusObjectClass
{
  GObjectClass parent;
};

#define NNTPGRAB_DBUS_TYPE_OBJECT              (nntpgrab_dbus_object_get_type ())
#define NNTPGRAB_DBUS_OBJECT(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), NNTPGRAB_DBUS_TYPE_OBJECT, NNTPGrabDBusObject))
#define NNTPGRAB_DBUS_OBJECT_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), NNTPGRAB_DBUS_TYPE_OBJECT, OpenFTDDBusObjectClass))
#define NNTPGRAB_DBUS_IS_OBJECT(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), NNTPGRAB_DBUS_TYPE_OBJECT))
#define NNTPGRAB_DBUS_IS_OBJECT_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), NNTPGRAB_DBUS_TYPE_OBJECT))
#define NNTPGRAB_DBUS_OBJECT_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), NNTPGRAB_DBUS_TYPE_OBJECT, NNTPGrabDBusObjectClass))

G_DEFINE_TYPE(NNTPGrabDBusObject, nntpgrab_dbus_object, G_TYPE_OBJECT)

gboolean nntpgrab_dbus_ping(NNTPGrabDBusObject *obj);
gboolean nntpgrab_dbus_open_nzb_file(NNTPGrabDBusObject *obj, const char *nzbfile);

void        do_open_nzb_file(const char *nzb_file);

#include "nntpgrab_gui_dbus_glue.h"

static void
nntpgrab_dbus_object_init (NNTPGrabDBusObject *obj)
{
}

static void
nntpgrab_dbus_object_class_init (NNTPGrabDBusObjectClass *klass)
{
}

static gboolean
test_is_already_active(DBusGConnection *connection, gboolean also_check_server)
{
    DBusGProxy *proxy = NULL;

    /* Create a proxy object for the "bus driver" (name "org.freedesktop.DBus") */
    proxy = dbus_g_proxy_new_for_name (connection,
                                       "org.NNTPGrab.NNTPGrab",
                                       "/org/NNTPGrab/NNTPGrab/object",
                                       "org.NNTPGrab.NNTPGrab");

    g_assert(proxy);

    if (dbus_g_proxy_call (proxy, "ping", NULL, G_TYPE_INVALID, G_TYPE_INVALID)) {
        // Call succeeded, another NNTPGrab process is active
        g_object_unref (proxy);
        return TRUE;
    }

    g_object_unref (proxy);

    if (!also_check_server) {
        return FALSE;
    }

    // Is the nntpgrab server running?
    if (glue && nntpgrab_glue_get_is_standalone(glue) && test_is_nntpgrab_server_already_running()) {
        nntpgrab_gui_base_dialog_show(nntpgrab_gui_base_get_widget("windowMain"), _("The NNTPGrab Server is already running.\nThis program and the NNTPGrab Server cannot run at the same time"), GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);
        return TRUE;
    }

    return FALSE;
}

gboolean
ipc_initialize(void)
{
    GError *error = NULL;
    guint request_name_result;

    bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
    if (!bus) {
        char *errmsg = g_strdup_printf("Unable to connect to DBus:\n%s\n", error->message);

        nntpgrab_gui_base_dialog_show(nntpgrab_gui_base_get_widget("windowMain"), errmsg, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);

        g_error_free(error);
        g_free(errmsg);
        return TRUE;
    }

    if (test_is_already_active(bus, TRUE)) {
        // Another NNTPGrab process already detected, we can shut down our process
        return FALSE;
    }

    dbus_g_object_type_install_info (NNTPGRAB_DBUS_TYPE_OBJECT, &dbus_glib_nntpgrab_dbus_object_info);

    bus_proxy = dbus_g_proxy_new_for_name (bus, "org.freedesktop.DBus",
                                                "/org/freedesktop/DBus",
                                                "org.freedesktop.DBus");

    if (!dbus_g_proxy_call (bus_proxy, "RequestName", &error,
                            G_TYPE_STRING, "org.NNTPGrab.NNTPGrab",
                            G_TYPE_UINT, 0,
                            G_TYPE_INVALID,
                            G_TYPE_UINT, &request_name_result,
                            G_TYPE_INVALID)) {
        char *errmsg = g_strdup_printf("Unable to register NNTPGrab on DBus:\n%s\n", error->message);

        nntpgrab_gui_base_dialog_show(nntpgrab_gui_base_get_widget("windowMain"), errmsg, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);

        g_error_free(error);
        g_free(errmsg);
    }

    obj = g_object_new (NNTPGRAB_DBUS_TYPE_OBJECT, NULL);
    dbus_g_connection_register_g_object (bus, "/org/NNTPGrab/NNTPGrab/object", G_OBJECT (obj));

    return TRUE;
}

void
ipc_cleanup(void)
{
    if (bus_proxy) {
        g_object_unref(bus_proxy);
        bus_proxy = NULL;
    }

    if (obj) {
        g_object_unref(obj);
    }

    if (bus) {
        dbus_g_connection_unref(bus);
        bus = NULL;
    }
}

gboolean nntpgrab_dbus_ping(NNTPGrabDBusObject *obj)
{
    return TRUE;
}

gboolean
nntpgrab_dbus_open_nzb_file(NNTPGrabDBusObject *obj, const char *nzbfile)
{
    do_open_nzb_file(nzbfile);
    return TRUE;
}

void
ipc_send_open_nzb_file_request(const char *nzb_file)
{
    DBusGConnection *connection;
    GError *error;
    DBusGProxy *proxy;

    error = NULL;
    connection = dbus_g_bus_get (DBUS_BUS_SESSION,
                                &error);
    if (connection == NULL) {
        g_printerr ("Failed to open connection to bus: %s\n", error->message);
        g_error_free (error);
        exit (1);
    }

    proxy = dbus_g_proxy_new_for_name (connection,
                                        "org.NNTPGrab.NNTPGrab",
                                        "/org/NNTPGrab/NNTPGrab/object",
                                        "org.NNTPGrab.NNTPGrab");

    error = NULL;
    if (!dbus_g_proxy_call (proxy, "open_nzb_file", &error, G_TYPE_STRING, nzb_file, G_TYPE_INVALID, G_TYPE_INVALID)) {
        if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION) {
            g_printerr ("Caught remote method exception %s: %s",
                    dbus_g_error_get_name (error),
                    error->message);
        } else {
            g_printerr ("Error: %s\n", error->message);
        }
        g_error_free (error);
        exit (1);
    }

    g_object_unref (proxy);
}

gboolean
ipc_test_is_nntpgrab_already_running(gboolean also_check_server)
{
    if (!bus) {
        GError *error = NULL;

        bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
        if (!bus) {
            char *errmsg = g_strdup_printf("Unable to connect to DBus:\n%s\n", error->message);

            nntpgrab_gui_base_dialog_show(nntpgrab_gui_base_get_widget("windowMain"), errmsg, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);

            g_error_free(error);
            g_free(errmsg);
            return TRUE;
        }
    }

    return test_is_already_active(bus, also_check_server);
}

#else

gboolean
ipc_initialize(void)
{
    return TRUE;
}

void
ipc_cleanup(void)
{
}

void
ipc_send_open_nzb_file_request(const char *nzb_file)
{
}

gboolean
ipc_test_is_nntpgrab_already_running(gboolean also_check_server)
{
    if (!also_check_server) {
        return FALSE;
    }

    // Is the nntpgrab server running?
    if (nntpgrab_glue_get_is_standalone(glue) && test_is_nntpgrab_server_already_running()) {
        nntpgrab_gui_base_dialog_show(nntpgrab_gui_base_get_widget("windowMain"), _("The NNTPGrab Server is already running.\nThis program and the NNTPGrab Server cannot run at the same time"), GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);
        return TRUE;
    }

    return FALSE;
}

#endif /* HAVE_DBUS */
