/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef _MSC_VER
#include "config.h.win32"
#else
#include "config.h"
#endif

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <glib-object.h>

#include "nntpgrab.h"
#include "nntpgrab_utils.h"
#include "nntpgrab_internal.h"
#include "collections.h"

static void
config_changed (NNTPGrabCore *core, gpointer data)
{
    nntpgrab_core_emit_config_changed(TRUE);
}

static void
part_download_start (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, gpointer data)
{
    nntpgrab_core_emit_part_download_start(TRUE, servername, conn_id, collection_name, subject, part_num);
}

static void
part_done (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gpointer data)
{
    nntpgrab_core_emit_part_done(TRUE, servername, conn_id, collection_name, subject, part_num, size);
}

static void
part_failed (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int size, gboolean all_servers_tried, gpointer data)
{
    nntpgrab_core_emit_part_failed(TRUE, servername, conn_id, collection_name, subject, part_num, size, all_servers_tried);
}

static void
traffic_monitor_update (NNTPGrabCore *core, int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, guint64 stamp, double average, gpointer data)
{
    int bytes_received[10];

    bytes_received[0] = bytes_received1;
    bytes_received[1] = bytes_received2;
    bytes_received[2] = bytes_received3;
    bytes_received[3] = bytes_received4;
    bytes_received[4] = bytes_received5;
    bytes_received[5] = bytes_received6;
    bytes_received[6] = bytes_received7;
    bytes_received[7] = bytes_received8;
    bytes_received[8] = bytes_received9;
    bytes_received[9] = bytes_received10;

    nntpgrab_core_emit_traffic_monitor_update(TRUE, bytes_received, stamp);
}

static void
part_progress_update (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num, int bytes_downloaded, int bytes_total, gpointer data)
{
    nntpgrab_core_emit_part_progress_update(TRUE, servername, conn_id, collection_name, subject, part_num, bytes_downloaded, bytes_total);
}

static void
collection_added (NNTPGrabCore *core, const char *collection_name, const char *poster, gpointer data)
{
    nntpgrab_core_emit_collection_added(TRUE, collection_name, poster);
}

static void
collection_removed (NNTPGrabCore *core, const char *collection_name, gpointer data)
{
    nntpgrab_core_emit_collection_removed(TRUE, collection_name);
}

static void
collection_modified (NNTPGrabCore *core, const char *collection_name, const char *poster, gpointer data)
{
    nntpgrab_core_emit_collection_modified(TRUE, collection_name, poster);
}

static void
file_added (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *poster, guint64 stamp, guint64 file_size, guint64 total_size, guint64 total_size_remaining, NGTaskState status, int num_parts, GList *groups, gpointer data)
{
    nntpgrab_core_emit_file_added(  TRUE, collection_name, subject, poster, stamp, file_size,
                                    total_size, total_size_remaining, status, num_parts, groups);
}

static void
file_removed (NNTPGrabCore *core, const char *collection_name, const char *subject, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    nntpgrab_core_emit_file_removed(TRUE, collection_name, subject, total_file_size, total_file_size_remaining);
}

static void
file_download_state_update (NNTPGrabCore *core, const char *collection_name, const char *subject, int num_parts_total, int num_parts_done, int num_parts_failed, guint64 file_size, guint64 file_size_remaining, guint64 total_file_size, guint64 total_file_size_remaining, gpointer data)
{
    nntpgrab_core_emit_file_download_state_update(TRUE, collection_name, subject, num_parts_total, num_parts_done, num_parts_failed, file_size, file_size_remaining, total_file_size, total_file_size_remaining);
}

static void
file_state_changed (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *real_filename, NGTaskState old_state, NGTaskState new_state, gpointer data)
{
    nntpgrab_core_emit_file_state_changed(TRUE, collection_name, subject, real_filename, old_state, new_state);
}

static void
connection_connecting (NNTPGrabCore *core, const char *servername, int conn_id, gpointer data)
{
    nntpgrab_core_emit_connecting(TRUE, servername, conn_id);
}

static void
connection_connected (NNTPGrabCore *core, const char *servername, int conn_id, const char *welcome_msg, gpointer data)
{
    nntpgrab_core_emit_connected(TRUE, servername, conn_id, welcome_msg);
}

static void
connection_disconnect (NNTPGrabCore *core, const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *msg, gpointer data)
{
    nntpgrab_core_emit_disconnect(TRUE, servername, conn_id, disconnect_type, msg);
}

static void
schedular_state_changed (NNTPGrabCore *core, NGSchedularState new_state, const char *reason, gpointer data)
{
    nntpgrab_core_emit_schedular_state_changed(TRUE, new_state, reason);
}

static void
log_message (NNTPGrabCore *core, const char *component, NGLogLevel log_level, const char *msg, gpointer data)
{
    nntpgrab_core_emit_log_message(TRUE, component, log_level, msg);
}

static void
task_moved (NNTPGrabCore *core, const char *orig_collection_name, const char *subject, const char *new_collection_name, int old_position, int new_position, gpointer data)
{
    nntpgrab_core_emit_file_moved(TRUE, orig_collection_name, subject, new_collection_name, old_position, new_position);
}

static void
collection_moved (NNTPGrabCore *core, const char *collection_name, int old_position, int new_position, gpointer data)
{
    nntpgrab_core_emit_collection_moved(TRUE, collection_name, old_position, new_position);
}

static void
all_downloads_completed (NNTPGrabCore *core, gpointer data)
{
    nntpgrab_core_emit_all_downloads_completed(TRUE);
}

static void
plugin_loaded (NNTPGrabCore *core, const char *plugin_name, gboolean is_persistent, gpointer data)
{
    nntpgrab_core_emit_plugin_loaded(TRUE, plugin_name, is_persistent);
}

static void
plugin_unloaded (NNTPGrabCore *core, const char *plugin_name, gpointer data)
{
    nntpgrab_core_emit_plugin_unloaded(TRUE, plugin_name);
}

static void
plugin_event (NNTPGrabCore *core, const char *plugin_name, const char *event_name, const char **values)
{
    nntpgrab_core_emit_plugin_event(TRUE, plugin_name, event_name, values);
}

void
connect_signal_handlers(NNTPGrabCore *core)
{
    gpointer data = NULL;

    ng_signal_connect(core, "config_changed", NG_CALLBACK(config_changed), data);
    ng_signal_connect(core, "part_download_start", NG_CALLBACK(part_download_start), data);
    ng_signal_connect(core, "part_done", NG_CALLBACK(part_done), data);
    ng_signal_connect(core, "part_failed", NG_CALLBACK(part_failed), data);
    ng_signal_connect(core, "traffic_monitor_update", NG_CALLBACK(traffic_monitor_update), data);
    ng_signal_connect(core, "part_progress_update", NG_CALLBACK(part_progress_update), data);
    ng_signal_connect(core, "collection_added", NG_CALLBACK(collection_added), data);
    ng_signal_connect(core, "collection_removed", NG_CALLBACK(collection_removed), data);
    ng_signal_connect(core, "collection_modified", NG_CALLBACK(collection_modified), data);
    ng_signal_connect(core, "file_added", NG_CALLBACK(file_added), data);
    ng_signal_connect(core, "file_removed", NG_CALLBACK(file_removed), data);
    ng_signal_connect(core, "file_download_state_update", NG_CALLBACK(file_download_state_update), data);
    ng_signal_connect(core, "file_state_changed", NG_CALLBACK(file_state_changed), data);
    ng_signal_connect(core, "connection_connecting", NG_CALLBACK(connection_connecting), data);
    ng_signal_connect(core, "connection_connected", NG_CALLBACK(connection_connected), data);
    ng_signal_connect(core, "connection_disconnect", NG_CALLBACK(connection_disconnect), data);
    ng_signal_connect(core, "schedular_state_changed", NG_CALLBACK(schedular_state_changed), data);
    ng_signal_connect(core, "log_message", NG_CALLBACK(log_message), data);
    ng_signal_connect(core, "task_moved", NG_CALLBACK(task_moved), data);
    ng_signal_connect(core, "collection_moved", NG_CALLBACK(collection_moved), data);
    ng_signal_connect(core, "all_downloads_completed", NG_CALLBACK(all_downloads_completed), data);
    ng_signal_connect(core, "plugin_loaded", NG_CALLBACK(plugin_loaded), data);
    ng_signal_connect(core, "plugin_unloaded", NG_CALLBACK(plugin_unloaded), data);
    ng_signal_connect(core, "plugin_event", NG_CALLBACK(plugin_event), data);
}

void
disconnect_signal_handlers(NNTPGrabCore *core)
{
    gpointer data = NULL;

    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(config_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_download_start), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_done), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_failed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(traffic_monitor_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(part_progress_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_added), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_removed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_modified), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_added), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_removed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_download_state_update), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(file_state_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(connection_connecting), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(connection_connected), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(connection_disconnect), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(schedular_state_changed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(log_message), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(task_moved), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(collection_moved), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(all_downloads_completed), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(plugin_loaded), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(plugin_unloaded), data);
    ng_signal_handlers_disconnect_by_func(core, NG_CALLBACK(plugin_event), data);
}
