/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n.h>

#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "decoder_thread.h"
#include "collection_alloc.h"
#include "download_thread.h"
#include "download_queue.h"
#include "nntpgrab_plugin.h"
#include "nntpgrab_utils.h"
#include "plugins.h"

static GStaticMutex mutex = G_STATIC_MUTEX_INIT;
static GThreadPool *decoderPool = NULL;
static gboolean abort_flag = FALSE;
static ConfigOpts opts;

static NNTPGrabDecoderRes
decoder_thread_decode_file(const char *collection_name, NNTPFile *file, int *saved_errno, char **errmsg)
{
    NNTPGrabDecoderRes retval;

    g_return_val_if_fail(collection_name != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(file != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(saved_errno != NULL, DECODER_RESULT_ERROR);
    g_return_val_if_fail(errmsg != NULL, DECODER_RESULT_ERROR);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("Trying to decode file with from collection '%s' with subject = '%s'"), collection_name, file->subject);

    g_return_val_if_fail(ng_plugin_call(NULL, "decode_file", collection_name, file, opts.temp_directory, opts.download_directory, saved_errno, errmsg, &retval), FALSE);

    ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, _("Decoder returned %i"), retval);

    if (retval == DECODER_RESULT_ERROR && errmsg && *errmsg) {
        ng_plugin_emit_log_msg(NULL, NG_LOG_LEVEL_INFO, "Decoder error: %s", *errmsg);
    }

    // The filename got updated, check the extension again
    if (retval != DECODER_RESULT_ERROR &&
        retval != DECODER_RESULT_NO_PARTS_AVAIL &&
        file->file_type == NNTP_FILE_TYPE_UNKNOWN) {

        file->file_type = nntpgrab_utils_get_file_type_of_filename(file->real_filename);
    }

    return retval;
}

static void
decoder_thread_func(gpointer data, gpointer user_data)
{
    DecoderData *decoder_data = (DecoderData *) data;
    int saved_errno = 0;
    static time_t last_timestamp = 0;
    NNTPGrabDecoderRes res;
    char *errmsg = NULL;

    // Do we still need to decode?
    g_static_mutex_lock(&decoder_data->collection->mutex);
    if (abort_flag || decoder_data->file->stop_flag) {
        decoder_data->file->now_decoding = FALSE;
        g_static_mutex_unlock(&decoder_data->collection->mutex);

        file_unref(decoder_data->file);
        collection_unref(decoder_data->collection);

        g_slice_free(DecoderData, decoder_data);

        return;
    }
    g_static_mutex_unlock(&decoder_data->collection->mutex);

    nntpgrab_core_emit_file_state_changed(FALSE, decoder_data->collection->collection_name, decoder_data->file->subject, NULL, decoder_data->file->status, TASK_STATE_DECODING);

    g_static_mutex_lock(&decoder_data->collection->mutex);
    decoder_data->file->status = TASK_STATE_DECODING;
    g_static_mutex_unlock(&decoder_data->collection->mutex);

    // Perform the actual decoding
    if ((res = decoder_thread_decode_file(decoder_data->collection->collection_name, decoder_data->file, &saved_errno, &errmsg)) == DECODER_RESULT_ERROR) {
        if (saved_errno == ENOSPC) {
            // No free space available, stop the schedular
            if (errmsg) {
                download_thread_abort_without_waiting(errmsg);
                g_free(errmsg);
            } else {
                download_thread_abort_without_waiting(strerror(saved_errno));
            }

            /* Skip any already queued decodes because of the error */
            abort_flag = TRUE;

            g_static_mutex_lock(&decoder_data->collection->mutex);
            decoder_data->file->now_decoding = FALSE;
            nntpgrab_core_emit_file_state_changed(FALSE, decoder_data->collection->collection_name, decoder_data->file->subject, NULL, decoder_data->file->status, TASK_STATE_WAITING_FOR_DECODE);
            decoder_data->file->status = TASK_STATE_WAITING_FOR_DECODE;
            g_static_mutex_unlock(&decoder_data->collection->mutex);

            file_unref(decoder_data->file);
            collection_unref(decoder_data->collection);

            g_slice_free(DecoderData, decoder_data);

            return;
        } else {
            // Unknown decode error occured, stop the schedular
            char *msg;

            if (errmsg) {
                msg = errmsg;
            } else {
                msg = g_strdup_printf(_("Unknown decode error occured, %s (%i)"), strerror(saved_errno), saved_errno);
            }
            download_thread_abort_without_waiting(msg);
            g_free(msg);

            /* Skip any already queued decodes because of the error */
            abort_flag = TRUE;

            g_static_mutex_lock(&decoder_data->collection->mutex);
            decoder_data->file->now_decoding = FALSE;
            nntpgrab_core_emit_file_state_changed(FALSE, decoder_data->collection->collection_name, decoder_data->file->subject, NULL, decoder_data->file->status, TASK_STATE_WAITING_FOR_DECODE);
            decoder_data->file->status = TASK_STATE_WAITING_FOR_DECODE;
            g_static_mutex_unlock(&decoder_data->collection->mutex);

            file_unref(decoder_data->file);
            collection_unref(decoder_data->collection);

            g_slice_free(DecoderData, decoder_data);

            return;
        }
    }

    // Mark the file as completed
    g_static_mutex_lock(&decoder_data->collection->mutex);

    if (!decoder_data->file->file_is_new) {
        decoder_data->file->file_is_changed = TRUE;
    }

    // Was the task restarted while the decoding was busy?
    if (decoder_data->file->now_decoding) {
        // Nope, mark the file decode as completed
        NGTaskState state;

        decoder_data->file->file_is_downloaded = TRUE;
        decoder_data->file->now_decoding = FALSE;

        switch(res) {
            case DECODER_RESULT_INCOMPLETE:
                state = TASK_STATE_FINISHED_INCOMPLETE;
                break;

            case DECODER_RESULT_NO_PARTS_AVAIL:
                state = TASK_STATE_FINISHED_NO_PARTS_AVAIL;
                break;

            case DECODER_RESULT_ERROR:
                state = TASK_STATE_WAITING_FOR_DECODE;
                break;

            default:
            case DECODER_RESULT_COMPLETE:
                state = TASK_STATE_FINISHED_COMPLETE;
                break;
        }

        nntpgrab_core_emit_file_state_changed(FALSE, decoder_data->collection->collection_name, decoder_data->file->subject, decoder_data->file->real_filename, decoder_data->file->status, state);
        decoder_data->file->status = state;
    }

    g_static_mutex_unlock(&decoder_data->collection->mutex);

    download_queue_test_is_collection_downloaded(decoder_data->collection);

    file_unref(decoder_data->file);
    collection_unref(decoder_data->collection);

    g_slice_free(DecoderData, decoder_data);

    // Save the download queue only when there has been some time between now and the previous time we've decoded something
    if (time(NULL) > last_timestamp + 60) {
        if (!download_queue_save(&errmsg)) {
            download_thread_abort_without_waiting(errmsg);
        }

        last_timestamp = time(NULL);
    }
}

void
decoder_thread_initialize(Configuration *config)
{
    g_return_if_fail(config != NULL);

    opts = configuration_get_opts(config);
}

void
decoder_thread_destroy(void)
{
}

gboolean
decoder_thread_start(void)
{
    g_static_mutex_lock(&mutex);
    if (decoderPool) {
        g_static_mutex_unlock(&mutex);
        g_return_val_if_fail(decoderPool == NULL, FALSE);
    }

    abort_flag = FALSE;
    decoderPool = g_thread_pool_new(decoder_thread_func, NULL, 1, FALSE, NULL);
    g_static_mutex_unlock(&mutex);

    return TRUE;
}

gboolean
decoder_thread_stop(void)
{
    g_return_val_if_fail(decoderPool != NULL, FALSE);

    g_static_mutex_lock(&mutex);
    abort_flag = TRUE;
    g_thread_pool_free(decoderPool, FALSE, TRUE);
    decoderPool = NULL;
    abort_flag = FALSE;
    g_static_mutex_unlock(&mutex);

    return TRUE;
}

void
decoder_thread_push_task(DecoderData *task)
{
    g_return_if_fail(decoderPool != NULL);
    g_return_if_fail(task != NULL);

    g_static_mutex_lock(&mutex);
    g_thread_pool_push(decoderPool, task, NULL);
    g_static_mutex_unlock(&mutex);
}
