/*
    Copyright (C) 2005-2010  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <glib/gi18n.h>
#include "nntpgrab_plugin.h"
#include "nntpgrab_automation.h"
#include "config.h"

static void on_nzb_imported(NNTPGrabAutoImport *auto_import, const char *filename, const char *collection_name, gpointer data);
static void import_nzb_file(NGPlugin *plugin_data, NNTPGrabNZB *nzbfile, const char *collection_name, char **warnings);
static void config_changed_cb(NGPlugin *plugin_data, gpointer data);

void
nntpgrab_plugin_initialize(NGPlugin *plugin_data)
{
    ng_plugin_set_name(plugin_data, "Auto-import");
    ng_plugin_set_version(plugin_data, PACKAGE_VERSION);
    ng_plugin_set_author(plugin_data, "Erik van Pienbroek");
    ng_plugin_set_url(plugin_data, "http://www.nntpgrab.nl");
    ng_plugin_set_description(plugin_data, "Monitor a specific directory for new NZB files and import these automatically");

    ng_plugin_set_required_event(plugin_data, "config_changed");

    ng_plugin_create_event(plugin_data, "nzb_imported", 2);
}

ngboolean
nntpgrab_plugin_load(NGPlugin *plugin_data, char **errmsg)
{
    plugin_data->priv = nntpgrab_automation_auto_import_new((ConfigGetOptsFunc) plugin_data->core_funcs.config_get_opts, (ImportNZBFileFunc) import_nzb_file, plugin_data);

    ng_plugin_connect_event(plugin_data, "config_changed", NG_PLUGIN_FUNCTION(config_changed_cb), NULL);
    config_changed_cb(plugin_data, NULL);

    ng_signal_connect(plugin_data->priv, "nzb_imported", NG_CALLBACK(on_nzb_imported), plugin_data);

    return TRUE;
}

ngboolean
nntpgrab_plugin_can_unload(NGPlugin *plugin_data, char **reason)
{
    return TRUE;
}

void
nntpgrab_plugin_unload(NGPlugin *plugin_data)
{
    g_return_if_fail(plugin_data != NULL);
    g_return_if_fail(plugin_data->priv != NULL);

    nntpgrab_automation_auto_import_destroy(plugin_data->priv);
    plugin_data->priv = NULL;

}

void
nntpgrab_plugin_destroy(NGPlugin *plugin_data)
{
}

int
nntpgrab_plugin_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

static void
on_nzb_imported(NNTPGrabAutoImport *auto_import, const char *filename, const char *collection_name, gpointer data)
{
    NGPlugin *plugin_data = (NGPlugin*) data;
    const char *params[3];

    params[0] = filename;
    params[1] = collection_name;
    params[2] = NULL;

    ng_plugin_emit_event(plugin_data, "nzb_imported", params);
}

static void
import_nzb_file(NGPlugin *plugin_data, NNTPGrabNZB *nzbfile, const char *collection_name, char **warnings)
{
    NGList *list;
    char *errmsg = NULL;

    g_return_if_fail(plugin_data != NULL);
    g_return_if_fail(plugin_data->priv != NULL);

    list = nzbfile->files;

    if (warnings) {
        *warnings = NULL;
    }

    while (list) {
        NNTPGrabNZBFile *file = list->data;

        if (!plugin_data->core_funcs.schedular_add_file_to_queue(collection_name, file->subject, file->poster, file->stamp, file->filesize, file->groups, file->segments, &errmsg)) {
            if (warnings) {
                if (*warnings) {
                    char *tmp = g_strdup_printf("%s\n%s", *warnings, errmsg);
                    g_free(*warnings);
                    *warnings = tmp;
                } else {
                    *warnings = g_strdup_printf(_("File could not be added to the download queue:\n%s"), errmsg);
                }
                g_free(errmsg);
            }
        }

        list = ng_list_next(list);
    }
}

static void
config_changed_cb(NGPlugin *plugin_data, gpointer data)
{
    static ConfigOpts prev_opts;
    static gboolean been_here = FALSE;
    ConfigOpts opts;
    char *errmsg = NULL;

    if (!been_here) {
        been_here = TRUE;
        memset(&prev_opts, 0, sizeof(ConfigOpts));
    }

    opts = plugin_data->core_funcs.config_get_opts();

    if (prev_opts.enable_auto_import != opts.enable_auto_import) {
        if (opts.enable_auto_import) {
#ifndef __APPLE__
            if (!nntpgrab_automation_enable_auto_import(plugin_data->priv, opts.auto_import_directory, &errmsg)) {
                ng_plugin_emit_log_msg(plugin_data, NG_LOG_LEVEL_WARNING, _("Unable to activate directory monitoring: %s"), errmsg);
                g_free(errmsg);
            }
#endif
        } else {
            nntpgrab_automation_disable_auto_import(plugin_data->priv);
        }
    }
}
