/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *  File-Roller
 *
 *  Copyright (C) 2001, 2003, 2008 Free Software Foundation, Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#ifdef WIN32
#include <windows.h>
#else
#include <sys/wait.h>
#endif
#include <unistd.h>
#include <glib.h>
#include "fr-process.h"
#include "fr-marshal.h"
#include "glib-utils.h"

#define REFRESH_RATE 100
#define BUFFER_SIZE 16384

#ifdef WIN32
#include <windows.h>
#include <io.h>

HANDLE my_pipein[2], my_pipeout[2], my_pipeerr[2];
char   my_popenmode = ' ';

static BOOL APIENTRY MyCreatePipeEx(
   OUT LPHANDLE lpReadPipe,
   OUT LPHANDLE lpWritePipe,
   IN LPSECURITY_ATTRIBUTES lpPipeAttributes,
   IN DWORD nSize,
   DWORD dwReadMode,
   DWORD dwWriteMode
)

/*++

Routine Description:

The CreatePipeEx API is used to create an anonymous pipe I/O device.
Unlike CreatePipe FILE_FLAG_OVERLAPPED may be specified for one or
both handles.
Two handles to the device are created.  One handle is opened for
reading and the other is opened for writing.  These handles may be
used in subsequent calls to ReadFile and WriteFile to transmit data
through the pipe.

Arguments:

lpReadPipe - Returns a handle to the read side of the pipe.  Data
may be read from the pipe by specifying this handle value in a
subsequent call to ReadFile.

lpWritePipe - Returns a handle to the write side of the pipe.  Data
may be written to the pipe by specifying this handle value in a
subsequent call to WriteFile.

lpPipeAttributes - An optional parameter that may be used to specify
the attributes of the new pipe.  If the parameter is not
specified, then the pipe is created without a security
descriptor, and the resulting handles are not inherited on
process creation.  Otherwise, the optional security attributes
are used on the pipe, and the inherit handles flag effects both
pipe handles.

nSize - Supplies the requested buffer size for the pipe.  This is
only a suggestion and is used by the operating system to
calculate an appropriate buffering mechanism.  A value of zero
indicates that the system is to choose the default buffering
scheme.

Return Value:

TRUE - The operation was successful.

FALSE/NULL - The operation failed. Extended error status is available
using GetLastError.

--*/
{
 static ULONG PipeSerialNumber = 1;
 HANDLE ReadPipeHandle, WritePipeHandle;
 DWORD dwError;
 CHAR PipeNameBuffer[ MAX_PATH ];

 //
 // Only one valid OpenMode flag - FILE_FLAG_OVERLAPPED
 //
 if ((dwReadMode | dwWriteMode) & (~FILE_FLAG_OVERLAPPED)) {
  SetLastError(ERROR_INVALID_PARAMETER);
  return FALSE;
 }

 //
 //  Set the default timeout to 120 seconds
 //

 if (nSize == 0) {
  nSize = 4096;
 }

 sprintf( PipeNameBuffer,
  "\\\\.\\Pipe\\NNTPGrabUnpackPipe.%08x.%08x",
  (unsigned int) GetCurrentProcessId(),
  (unsigned int) PipeSerialNumber++
  );

 ReadPipeHandle = CreateNamedPipeA(
  PipeNameBuffer,
  PIPE_ACCESS_INBOUND | dwReadMode,
  PIPE_TYPE_BYTE | PIPE_NOWAIT,
  1,             // Number of pipes
  nSize,         // Out buffer size
  nSize,         // In buffer size
  120 * 1000,    // Timeout in ms
  lpPipeAttributes
  );

 if (! ReadPipeHandle) {
  return FALSE;
 }

 WritePipeHandle = CreateFileA(
  PipeNameBuffer,
  GENERIC_WRITE,
  0,                         // No sharing
  lpPipeAttributes,
  OPEN_EXISTING,
  FILE_ATTRIBUTE_NORMAL | dwWriteMode,
  NULL                       // Template file
  );

 if (INVALID_HANDLE_VALUE == WritePipeHandle) {
  dwError = GetLastError();
  CloseHandle( ReadPipeHandle );
  SetLastError(dwError);
  return FALSE;
 }

 *lpReadPipe = ReadPipeHandle;
 *lpWritePipe = WritePipeHandle;
 return( TRUE );
}

static int
my_pipe(HANDLE *readwrite)
{
    SECURITY_ATTRIBUTES sa;

    sa.nLength = sizeof(sa);
    sa.bInheritHandle = 1;
    sa.lpSecurityDescriptor = NULL;

#if 0
    if (!CreatePipe (&readwrite[0],&readwrite[1],&sa,1 << 13)) {
#else
    if (!MyCreatePipeEx(&readwrite[0],&readwrite[1],&sa,1 << 13, 0, FILE_FLAG_OVERLAPPED)) {
#endif
        errno = EMFILE;
        return -1;
    }

    return 0;
}

static FILE *
win32_popen(const char *cmd, const char *mode, HANDLE *hProcess, HANDLE *hPipe)
{
    FILE *fptr = (FILE *)0;
    PROCESS_INFORMATION piProcInfo;
    STARTUPINFO siStartInfo;
    int success, catch_stderr;

    my_pipein[0]   = INVALID_HANDLE_VALUE;
    my_pipein[1]   = INVALID_HANDLE_VALUE;
    my_pipeout[0]  = INVALID_HANDLE_VALUE;
    my_pipeout[1]  = INVALID_HANDLE_VALUE;
    my_pipeerr[0]  = INVALID_HANDLE_VALUE;
    my_pipeerr[1]  = INVALID_HANDLE_VALUE;

    if (!mode || !*mode)
        goto finito;

    my_popenmode = *mode;
    if (my_popenmode != 'r' && my_popenmode != 'w')
        goto finito;

    catch_stderr = strstr((char *) cmd, "2>&1") != NULL;
    if (catch_stderr) {
        char *ptr = strstr((char*) cmd, "2>&1");
        *ptr = '\0';
    }

    if (my_pipe(my_pipein)  == -1 || my_pipe(my_pipeout) == -1)
        goto finito;
    if (!catch_stderr && my_pipe(my_pipeerr) == -1)
      goto finito;

    ZeroMemory(&siStartInfo, sizeof(STARTUPINFO));
    siStartInfo.cb             = sizeof(STARTUPINFO);
    siStartInfo.hStdInput      = my_pipein[0];
    siStartInfo.hStdOutput     = my_pipeout[1];
    if (catch_stderr)
        siStartInfo.hStdError  = my_pipeout[1];
    else
        siStartInfo.hStdError  = my_pipeerr[1];
    siStartInfo.dwFlags        = STARTF_USESTDHANDLES;

    success = CreateProcess(NULL,
       (LPTSTR)cmd,       // command line
       NULL,              // process security attributes
       NULL,              // primary thread security attributes
       TRUE,              // handles are inherited
       DETACHED_PROCESS,  // creation flags
       NULL,              // use parent's environment
       NULL,              // use parent's current directory
       &siStartInfo,      // STARTUPINFO pointer
       &piProcInfo);      // receives PROCESS_INFORMATION

    if (!success)
        goto finito;

    CloseHandle(my_pipein[0]);  my_pipein[0]  = INVALID_HANDLE_VALUE;
    CloseHandle(my_pipeout[1]); my_pipeout[1] = INVALID_HANDLE_VALUE;
    CloseHandle(my_pipeerr[1]); my_pipeerr[1] = INVALID_HANDLE_VALUE;

    if (my_popenmode == 'r') {
        fptr = _fdopen(_open_osfhandle((long)my_pipeout[0],_O_BINARY),"r");
        if (hPipe) {
            *hPipe = my_pipeout[0];
        }
    } else {
        fptr = _fdopen(_open_osfhandle((long)my_pipein[1],_O_BINARY),"w");
        if (hPipe) {
            *hPipe = my_pipein[1];
        }
    }

    if (hProcess)
        *hProcess = piProcInfo.hProcess;

finito:
    if (!fptr) {
        if (my_pipein[0]  != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipein[0]);
        if (my_pipein[1]  != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipein[1]);
        if (my_pipeout[0] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeout[0]);
        if (my_pipeout[1] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeout[1]);
        if (my_pipeerr[0] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeerr[0]);
        if (my_pipeerr[1] != INVALID_HANDLE_VALUE)
            CloseHandle(my_pipeerr[1]);
    }

    return fptr;
}

static int
win32_pclose(FILE *file, HANDLE hProcess)
{
    if (!file) {
        return -1;
    }

    fclose(file);

    CloseHandle(my_pipeerr[0]);
    if (my_popenmode == 'r')
        CloseHandle(my_pipein[1]);
    else
       CloseHandle(my_pipeout[0]);

    if (hProcess)
        CloseHandle(hProcess);

    return 0;
}
#endif

enum {
	START,
	DONE,
	STICKY_ONLY,
	LAST_SIGNAL
};

static GObjectClass *parent_class;
static guint fr_process_signals[LAST_SIGNAL] = { 0 };

static void fr_process_class_init (FrProcessClass *class);
static void fr_process_init       (FrProcess      *process);
static void fr_process_finalize   (GObject        *object);


typedef struct {
	GList        *args;              /* command to execute */
	char         *dir;               /* working directory */
	guint         sticky : 1;        /* whether the command must be
					  * executed even if a previous
					  * command has failed. */
	guint         ignore_error : 1;  /* whether to continue to execute
					  * other commands if this command
					  * fails. */
	ContinueFunc  continue_func;
	gpointer      continue_data;
	ProcFunc      begin_func;
	gpointer      begin_data;
	ProcFunc      end_func;
	gpointer      end_data;
} FrCommandInfo;


static FrCommandInfo *
fr_command_info_new (void)
{
	FrCommandInfo *info;

	info = g_new0 (FrCommandInfo, 1);
	info->args = NULL;
	info->dir = NULL;
	info->sticky = FALSE;
	info->ignore_error = FALSE;

	return info;
}


static void
fr_command_info_free (FrCommandInfo *info)
{
	if (info == NULL)
		return;

	if (info->args != NULL) {
		g_list_foreach (info->args, (GFunc) g_free, NULL);
		g_list_free (info->args);
		info->args = NULL;
	}

	if (info->dir != NULL) {
		g_free (info->dir);
		info->dir = NULL;
	}

	g_free (info);
}


static void
fr_channel_data_init (FrChannelData *channel)
{
	channel->source_fd = NULL;
	channel->raw = NULL;
	channel->status = G_IO_STATUS_NORMAL;
	channel->error = NULL;
    channel->recv_buffer_len = 0;
    memset(channel->recv_buffer, 0, sizeof(channel->recv_buffer));
}

static void
fr_channel_data_close_source (FrChannelData *channel)
{
	if (channel->source_fd != NULL) {
#ifdef WIN32
        win32_pclose(channel->source_fd, NULL);
#else
        pclose(channel->source_fd);
#endif
        channel->source_fd = NULL;
	}
}

static GIOStatus
fr_channel_data_read (FrChannelData *channel)
{
	channel->status = G_IO_STATUS_NORMAL;
	g_clear_error (&channel->error);

	while (TRUE) {
#ifdef WIN32
		/* Win32 doesn't support async reads on pipes in combination with the fgets() function.. */
		char buf[2];
		DWORD num_read = 0;
		while (ReadFile(channel->hPipe, &buf, 1, &num_read, NULL)) {
			if (buf[0] == '\r' || (buf[0] == '\n' && channel->recv_buffer_len > 0)) {
				channel->raw = g_list_prepend (channel->raw, g_strdup(channel->recv_buffer));
				if (channel->line_func != NULL)
					(*channel->line_func) (channel->recv_buffer, channel->line_data);

				memset(channel->recv_buffer, 0, sizeof(channel->recv_buffer));
				channel->recv_buffer_len = 0;
			} else if (buf[0] == '\n') {
				/* Ignore */
			} else {
				channel->recv_buffer[channel->recv_buffer_len] = buf[0];
				channel->recv_buffer_len++;
			}
		}

		break;
#else
		if (!fgets(channel->recv_buffer, sizeof(channel->recv_buffer) - 1, channel->source_fd)) {
			break;
		}

		if (strlen(channel->recv_buffer) > 0) {
			if (channel->recv_buffer[strlen(channel->recv_buffer) - 1] == '\n') {
				channel->recv_buffer[strlen(channel->recv_buffer) - 1] = '\0';
			}

			if (channel->recv_buffer[strlen(channel->recv_buffer) - 1] == '\r') {
				channel->recv_buffer[strlen(channel->recv_buffer) - 1] = '\0';
			}
		}

		channel->raw = g_list_prepend (channel->raw, g_strdup(channel->recv_buffer));
		if (channel->line_func != NULL)
			(*channel->line_func) (channel->recv_buffer, channel->line_data);

		memset(channel->recv_buffer, 0, sizeof(channel->recv_buffer));
		channel->recv_buffer_len = 0;
#endif
	}

	/* Do we have have an EOF or some other error? */
	if (feof(channel->source_fd)) {
		channel->status = G_IO_STATUS_EOF;
	} else if (ferror(channel->source_fd) && errno != 0) {
#ifdef WIN32
		switch (GetLastError()) {
			case ERROR_NO_DATA:
				channel->status = G_IO_STATUS_AGAIN;
				break;

			default:
				channel->status = G_IO_STATUS_ERROR;
				break;
		}
#else
		switch (errno) {
			case EAGAIN:
				channel->status = G_IO_STATUS_AGAIN;
				break;

			default:
				channel->status = G_IO_STATUS_ERROR;
				break;
		}
#endif
	}

	return channel->status;
}


#if 0
static GIOStatus
fr_channel_data_flush (FrChannelData *channel)
{
	GIOStatus status;

	while (((status = fr_channel_data_read (channel)) != G_IO_STATUS_ERROR) && (status != G_IO_STATUS_EOF))
		/* void */;
	fr_channel_data_close_source (channel);

	return status;
}
#endif


static void
fr_channel_data_reset (FrChannelData *channel)
{
	fr_channel_data_close_source (channel);

	if (channel->raw != NULL) {
		g_list_foreach (channel->raw, (GFunc) g_free, NULL);
		g_list_free (channel->raw);
		channel->raw = NULL;
	}
}


static void
fr_channel_data_free (FrChannelData *channel)
{
	fr_channel_data_reset (channel);
}


static void
fr_channel_data_set_fd (FrChannelData *channel,
			FILE          *fd,
			const char    *charset)
{
#ifdef WIN32
    u_long          ioctlArg;
#else
    int             mode;
#endif

	fr_channel_data_reset (channel);

	channel->source_fd = fd;

	/* Set the pipe non-blocking */
#ifdef WIN32
    ioctlArg = 1;
    ioctlsocket(fileno(channel->source_fd), FIONBIO, &ioctlArg);
#else
    mode = fcntl(fileno(channel->source_fd), F_GETFL, 0);
    fcntl(fileno(channel->source_fd), F_SETFL, mode | O_NONBLOCK);
#endif

    /* Disable buffering */
    setvbuf(channel->source_fd, (char *) NULL, _IOLBF, 0);

#if 0
	g_io_channel_set_buffer_size (channel->source, BUFFER_SIZE);
	if (charset != NULL)
		g_io_channel_set_encoding (channel->source, charset, NULL);
#endif
}


const char *try_charsets[] = { "UTF-8", "ISO-8859-1", "WINDOW-1252" };
int n_charsets = G_N_ELEMENTS (try_charsets);


struct _FrProcessPrivate {
	GPtrArray   *comm;                /* FrCommandInfo elements. */
	gint         n_comm;              /* total number of commands */
	gint         current_comm;        /* currenlty editing command. */

	GPid         command_pid;
	guint        check_timeout;

	FrProcError  first_error;

	gboolean     running;
	gboolean     stopping;
	gint         current_command;
	gint         error_command;       /* command that coused an error. */

	gboolean     use_standard_locale;
	gboolean     sticky_only;         /* whether to execute only sticky
			 		   * commands. */
	int          current_charset;
};


GType
fr_process_get_type (void)
{
	static GType type = 0;

	if (! type) {
		GTypeInfo type_info = {
			sizeof (FrProcessClass),
			NULL,
			NULL,
			(GClassInitFunc) fr_process_class_init,
			NULL,
			NULL,
			sizeof (FrProcess),
			0,
			(GInstanceInitFunc) fr_process_init
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "FRProcess",
					       &type_info,
					       0);
	}

	return type;
}


static void
fr_process_class_init (FrProcessClass *class)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (class);

	parent_class = g_type_class_peek_parent (class);

	fr_process_signals[START] =
		g_signal_new ("start",
			      G_TYPE_FROM_CLASS (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (FrProcessClass, start),
			      NULL, NULL,
			      fr_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	fr_process_signals[DONE] =
		g_signal_new ("done",
			      G_TYPE_FROM_CLASS (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (FrProcessClass, done),
			      NULL, NULL,
			      fr_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);
	fr_process_signals[STICKY_ONLY] =
		g_signal_new ("sticky_only",
			      G_TYPE_FROM_CLASS (class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (FrProcessClass, sticky_only),
			      NULL, NULL,
			      fr_marshal_VOID__VOID,
			      G_TYPE_NONE, 0);

	gobject_class->finalize = fr_process_finalize;

	class->start = NULL;
	class->done  = NULL;
}


static void
fr_process_init (FrProcess *process)
{
	process->priv = g_new0 (FrProcessPrivate, 1);

	process->term_on_stop = TRUE;

	process->priv->comm = g_ptr_array_new ();
	process->priv->n_comm = -1;
	process->priv->current_comm = -1;

#ifdef WIN32
    process->priv->command_pid = INVALID_HANDLE_VALUE;
#else
	process->priv->command_pid = 0;
#endif

	fr_channel_data_init (&process->out);
#if 0
	fr_channel_data_init (&process->err);
#endif

	process->error.gerror = NULL;
	process->priv->first_error.gerror = NULL;

	process->priv->check_timeout = 0;
	process->priv->running = FALSE;
	process->priv->stopping = FALSE;
	process->restart = FALSE;

	process->priv->current_charset = -1;

	process->priv->use_standard_locale = FALSE;
}


FrProcess *
fr_process_new (void)
{
	return FR_PROCESS (g_object_new (FR_TYPE_PROCESS, NULL));
}


static void fr_process_stop_priv (FrProcess *process, gboolean emit_signal);


static void
fr_process_finalize (GObject *object)
{
	FrProcess *process;

	g_return_if_fail (object != NULL);
	g_return_if_fail (FR_IS_PROCESS (object));

	process = FR_PROCESS (object);

	fr_process_stop_priv (process, FALSE);
	fr_process_clear (process);

	g_ptr_array_free (process->priv->comm, FALSE);

	fr_channel_data_free (&process->out);
#if 0
	fr_channel_data_free (&process->err);
#endif

	g_clear_error (&process->error.gerror);
	g_clear_error (&process->priv->first_error.gerror);

	g_free (process->priv);

	/* Chain up */

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (object);
}


void
fr_process_begin_command (FrProcess  *process,
			  const char *arg)
{
	FrCommandInfo *info;
#ifdef WIN32
	char *path;
#endif

	g_return_if_fail (process != NULL);

	info = fr_command_info_new ();
#ifdef WIN32
	path = g_find_program_in_path(arg);
	if (path) {
		info->args = g_list_prepend (NULL, path);
	} else {
		info->args = g_list_prepend (NULL, g_strdup_printf("%s.exe", arg));
	}
#else
	info->args = g_list_prepend (NULL, g_strdup (arg));
#endif

	g_ptr_array_add (process->priv->comm, info);

	process->priv->n_comm++;
	process->priv->current_comm = process->priv->n_comm;
}


void
fr_process_begin_command_at (FrProcess  *process,
			     const char *arg,
			     int         index)
{
	FrCommandInfo *info, *old_c_info;

	g_return_if_fail (process != NULL);
	g_return_if_fail (index >= 0 && index <= process->priv->n_comm);

	process->priv->current_comm = index;

	old_c_info = g_ptr_array_index (process->priv->comm, index);

	if (old_c_info != NULL)
		fr_command_info_free (old_c_info);

	info = fr_command_info_new ();
	info->args = g_list_prepend (NULL, g_strdup (arg));

	g_ptr_array_index (process->priv->comm, index) = info;
}


void
fr_process_set_working_dir (FrProcess  *process,
			    const char *dir)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);
	g_return_if_fail (process->priv->current_comm >= 0);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	if (info->dir != NULL)
		g_free (info->dir);
	info->dir = g_strdup (dir);
}


void
fr_process_set_sticky (FrProcess *process,
		       gboolean   sticky)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);
	g_return_if_fail (process->priv->current_comm >= 0);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->sticky = sticky;
}


void
fr_process_set_ignore_error (FrProcess *process,
			     gboolean   ignore_error)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);
	g_return_if_fail (process->priv->current_comm >= 0);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->ignore_error = ignore_error;
}


void
fr_process_add_arg (FrProcess  *process,
		    const char *arg)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);
	g_return_if_fail (process->priv->current_comm >= 0);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->args = g_list_prepend (info->args, g_strdup (arg));
}


void
fr_process_add_arg_concat (FrProcess  *process,
			   const char *arg1,
			   ...)
{
	GString *arg;
	va_list  args;
	char    *s;

	arg = g_string_new (arg1);

	va_start (args, arg1);
	while ((s = va_arg (args, char*)) != NULL)
		g_string_append (arg, s);
	va_end (args);

	fr_process_add_arg (process, arg->str);
	g_string_free (arg, TRUE);
}


void
fr_process_add_arg_printf (FrProcess    *fr_proc,
			   const char   *format,
			   ...)
{
	va_list  args;
	char    *arg;

	va_start (args, format);
	arg = g_strdup_vprintf (format, args);
	va_end (args);

	fr_process_add_arg (fr_proc, arg);

	g_free (arg);
}


void
fr_process_set_arg_at (FrProcess  *process,
		       int         n_comm,
		       int         n_arg,
		       const char *arg_value)
{
	FrCommandInfo *info;
	GList         *arg;

	g_return_if_fail (process != NULL);

	info = g_ptr_array_index (process->priv->comm, n_comm);
	arg = g_list_nth (info->args, n_arg);
	g_return_if_fail (arg != NULL);

	g_free (arg->data);
	arg->data = g_strdup (arg_value);
}


void
fr_process_set_begin_func (FrProcess    *process,
			   ProcFunc      func,
			   gpointer      func_data)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->begin_func = func;
	info->begin_data = func_data;
}


void
fr_process_set_end_func (FrProcess    *process,
			 ProcFunc      func,
			 gpointer      func_data)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->end_func = func;
	info->end_data = func_data;
}


void
fr_process_set_continue_func (FrProcess    *process,
			      ContinueFunc  func,
			      gpointer      func_data)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);

	if (process->priv->current_comm < 0)
		return;

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->continue_func = func;
	info->continue_data = func_data;
}


void
fr_process_end_command (FrProcess *process)
{
	FrCommandInfo *info;

	g_return_if_fail (process != NULL);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_comm);
	info->args = g_list_reverse (info->args);
}


void
fr_process_clear (FrProcess *process)
{
	gint i;

	g_return_if_fail (process != NULL);

	for (i = 0; i <= process->priv->n_comm; i++) {
		FrCommandInfo *info;

		info = g_ptr_array_index (process->priv->comm, i);
		fr_command_info_free (info);
		g_ptr_array_index (process->priv->comm, i) = NULL;
	}

	for (i = 0; i <= process->priv->n_comm; i++)
		g_ptr_array_remove_index_fast (process->priv->comm, 0);

	process->priv->n_comm = -1;
	process->priv->current_comm = -1;
}


void
fr_process_set_out_line_func (FrProcess *process,
			      LineFunc   func,
			      gpointer   data)
{
	g_return_if_fail (process != NULL);

	process->out.line_func = func;
	process->out.line_data = data;
}


void
fr_process_set_err_line_func (FrProcess *process,
			      LineFunc   func,
			      gpointer   data)
{
	g_return_if_fail (process != NULL);

#if 0
	process->err.line_func = func;
	process->err.line_data = data;
#endif
}


static gboolean check_child (gpointer data);


#if 0
static void child_setup (gpointer user_data)
{
	FrProcess *process = user_data;

	if (process->priv->use_standard_locale)
		putenv ("LC_MESSAGES=C");
}
#endif


static const char *
fr_process_get_charset (FrProcess *process)
{
	const char *charset = NULL;

	if (process->priv->current_charset >= 0)
		charset = try_charsets[process->priv->current_charset];
	else if (g_get_charset (&charset))
		charset = NULL;

	return charset;
}

static void
start_current_command (FrProcess *process)
{
	FrCommandInfo  *info;
	GList          *scan;
	char           *cmd = NULL;
	FILE           *out_fd;
#if 0
    FILE           *err_fd;
#endif
    char           *tmp;

	debug (DEBUG_INFO, "%d/%d) ", process->priv->current_command, process->priv->n_comm);

	info = g_ptr_array_index (process->priv->comm, process->priv->current_command);

	for (scan = info->args; scan; scan = scan->next) {
        if (cmd == NULL) {
            cmd = g_strdup_printf("%s", (const char*) scan->data);
        } else {
            tmp = g_strdup_printf("%s \"%s\"", cmd, (const char*) scan->data);
            g_free(cmd);
            cmd = tmp;
        }
    }

    tmp = g_strdup_printf("%s 2>&1 < /dev/null", cmd);
    g_free(cmd);
    cmd = tmp;

#ifdef DEBUG
	{
		if (process->priv->use_standard_locale)
			g_print ("\tLC_MESSAGES=C\n");

		if (info->dir != NULL)
			g_print ("\tcd %s\n", info->dir);

		g_print ("\t");
		g_print ("%s\n", cmd);
		g_print ("\n");
	}
#endif

	if (info->begin_func != NULL)
		(*info->begin_func) (info->begin_data);

#ifdef WIN32
    out_fd = win32_popen(cmd, "r", &process->priv->command_pid, &process->out.hPipe);
#else
    out_fd = popen(cmd, "r");
#endif
    if (!out_fd) {
        process->error.type = FR_PROC_ERROR_SPAWN;
        g_signal_emit (G_OBJECT (process),
                   fr_process_signals[DONE],
                   0);
        g_free (cmd);
        return;
    }

	g_free(cmd);

	fr_channel_data_set_fd (&process->out, out_fd, fr_process_get_charset (process));
#if 0
	fr_channel_data_set_fd (&process->err, err_fd, fr_process_get_charset (process));
#endif

	process->priv->check_timeout = g_timeout_add (REFRESH_RATE,
					              check_child,
					              process);
}


static gboolean
command_is_sticky (FrProcess *process,
		   int        i)
{
	FrCommandInfo *info;

	info = g_ptr_array_index (process->priv->comm, i);
	return info->sticky;
}


static void
allow_sticky_processes_only (FrProcess *process,
			     gboolean   emit_signal)
{
	if (! process->priv->sticky_only) {
		/* Remember the first error. */
		process->priv->error_command = process->priv->current_command;
		process->priv->first_error.type = process->error.type;
		process->priv->first_error.status = process->error.status;
		g_clear_error (&process->priv->first_error.gerror);
		if (process->error.gerror != NULL)
			process->priv->first_error.gerror = g_error_copy (process->error.gerror);
	}

	process->priv->sticky_only = TRUE;
	if (emit_signal)
		g_signal_emit (G_OBJECT (process),
			       fr_process_signals[STICKY_ONLY],
			       0);
}


static void
fr_process_set_error (FrProcess       *process,
		      FrProcErrorType  type,
		      int              status,
		      GError          *gerror)
{
	process->error.type = type;
	process->error.status = status;
	if (gerror != process->error.gerror) {
		g_clear_error (&process->error.gerror);
		if (gerror != NULL)
			process->error.gerror = g_error_copy (gerror);
	}
}


static gint
check_child (gpointer data)
{
	FrProcess      *process = data;
	FrCommandInfo  *info;
	gboolean        continue_process;
	gboolean        channel_error = FALSE;
#ifdef WIN32
	DWORD		exit_code = -1;
    GPid            pid;
#else
    pid_t           pid;
    int             status;
#endif

	info = g_ptr_array_index (process->priv->comm, process->priv->current_command);

	/* Remove check. */

	g_source_remove (process->priv->check_timeout);
	process->priv->check_timeout = 0;

	if (fr_channel_data_read (&process->out) == G_IO_STATUS_ERROR) {
		fr_process_set_error (process, FR_PROC_ERROR_IO_CHANNEL, 0, process->out.error);
		channel_error = TRUE;
	}
#if 0
	else if (fr_channel_data_read (&process->err) == G_IO_STATUS_ERROR) {
		fr_process_set_error (process, FR_PROC_ERROR_IO_CHANNEL, 0, process->err.error);
		channel_error = TRUE;
	}
#endif
	else {
#ifdef WIN32
		if (!GetExitCodeProcess(process->priv->command_pid, &exit_code)) {
			pid = process->priv->command_pid;
		} else {
			if (exit_code != STILL_ACTIVE) {
				/* Process has completed */
				pid = process->priv->command_pid;
			} else {
				pid = 0;
			}
		}
#else
		pid = waitpid (/* process->priv->command_pid */ -1, &status, WNOHANG);
#endif
		if (pid == 0) {
			/* Add check again. */
			process->priv->check_timeout = g_timeout_add (REFRESH_RATE,
							              check_child,
							              process);
			return FALSE;
		}
	}

	if (info->ignore_error) {
		process->error.type = FR_PROC_ERROR_NONE;
		debug (DEBUG_INFO, "[ignore error]\n");
	}
	else if (! channel_error && (process->error.type != FR_PROC_ERROR_STOPPED)) {
#ifdef WIN32
		if (exit_code == 0)
			process->error.type = FR_PROC_ERROR_NONE;
#else
		if (WIFEXITED (status)) {
			if (WEXITSTATUS (status) == 0)
				process->error.type = FR_PROC_ERROR_NONE;
			else if (WEXITSTATUS (status) == 255)
				process->error.type = FR_PROC_ERROR_COMMAND_NOT_FOUND;
			else {
				process->error.type = FR_PROC_ERROR_COMMAND_ERROR;
				process->error.status = WEXITSTATUS (status);
			}
		}
#endif
		else
			process->error.type = FR_PROC_ERROR_EXITED_ABNORMALLY;
	}

	process->priv->command_pid = 0;

#if 0
	if (fr_channel_data_flush (&process->out) == G_IO_STATUS_ERROR) {
		fr_process_set_error (process, FR_PROC_ERROR_IO_CHANNEL, 0, process->out.error);
		channel_error = TRUE;
	}
	else if (fr_channel_data_flush (&process->err) == G_IO_STATUS_ERROR) {
		fr_process_set_error (process, FR_PROC_ERROR_IO_CHANNEL, 0, process->err.error);
		channel_error = TRUE;
	}
#endif

	if (info->end_func != NULL)
		(*info->end_func) (info->end_data);

	/**/

	if (channel_error
	    && (process->error.type == FR_PROC_ERROR_IO_CHANNEL)
	    && g_error_matches (process->error.gerror, G_CONVERT_ERROR, G_CONVERT_ERROR_ILLEGAL_SEQUENCE))
	{
		if (process->priv->current_charset < n_charsets - 1) {
			/* try with another charset */
			process->priv->current_charset++;
			process->priv->running = FALSE;
			process->restart = TRUE;
			fr_process_start (process);
			return FALSE;
		}
		/*fr_process_set_error (process, FR_PROC_ERROR_NONE, 0, NULL);*/
		fr_process_set_error (process, FR_PROC_ERROR_BAD_CHARSET, 0, process->error.gerror);
	}

	/* Check whether to continue or stop the process */

	continue_process = TRUE;
	if (info->continue_func != NULL)
		continue_process = (*info->continue_func) (info->continue_data);

	/* Execute next command. */
	if (continue_process) {
		if (process->error.type != FR_PROC_ERROR_NONE)
			allow_sticky_processes_only (process, TRUE);

		if (process->priv->sticky_only) {
			do {
				process->priv->current_command++;
			} while ((process->priv->current_command <= process->priv->n_comm)
				 && ! command_is_sticky (process, process->priv->current_command));
		}
		else
			process->priv->current_command++;

		if (process->priv->current_command <= process->priv->n_comm) {
			start_current_command (process);
			return FALSE;
		}
	}

	/* Done */

	process->priv->current_command = -1;
	process->priv->use_standard_locale = FALSE;

	if (process->out.raw != NULL)
		process->out.raw = g_list_reverse (process->out.raw);
#if 0
	if (process->err.raw != NULL)
		process->err.raw = g_list_reverse (process->err.raw);
#endif

	process->priv->running = FALSE;
	process->priv->stopping = FALSE;

	if (process->priv->sticky_only) {
		/* Restore the first error. */
		fr_process_set_error (process,
				      process->priv->first_error.type,
				      process->priv->first_error.status,
				      process->priv->first_error.gerror);
	}

	g_signal_emit (G_OBJECT (process),
		       fr_process_signals[DONE],
		       0);

	return FALSE;
}


void
fr_process_use_standard_locale (FrProcess *process,
				gboolean   use_stand_locale)
{
	g_return_if_fail (process != NULL);
	process->priv->use_standard_locale = use_stand_locale;
}


void
fr_process_start (FrProcess *process)
{
	g_return_if_fail (process != NULL);

	if (process->priv->running)
		return;

	fr_channel_data_reset (&process->out);
#if 0
	fr_channel_data_reset (&process->err);
#endif

	process->priv->sticky_only = FALSE;
	process->priv->current_command = 0;
	fr_process_set_error (process, FR_PROC_ERROR_NONE, 0, NULL);

	if (! process->restart) {
		process->priv->current_charset = -1;
		g_signal_emit (G_OBJECT (process),
			       fr_process_signals[START],
			       0);
	}

	process->priv->stopping = FALSE;

	if (process->priv->n_comm == -1) {
		process->priv->running = FALSE;
		g_signal_emit (G_OBJECT (process),
			       fr_process_signals[DONE],
			       0);
	}
	else {
		process->priv->running = TRUE;
		start_current_command (process);
	}
}


static void
fr_process_stop_priv (FrProcess *process,
		      gboolean   emit_signal)
{
	g_return_if_fail (process != NULL);

	if (! process->priv->running)
		return;

	if (process->priv->stopping)
		return;

	process->priv->stopping = TRUE;
	process->error.type = FR_PROC_ERROR_STOPPED;

	if (command_is_sticky (process, process->priv->current_command))
		allow_sticky_processes_only (process, emit_signal);

	else if (process->term_on_stop)
#ifdef WIN32
		TerminateProcess(process->priv->command_pid, -1);
#else
		//kill (process->priv->command_pid, SIGTERM);
        ;
#endif

	else {
		if (process->priv->check_timeout != 0) {
			g_source_remove (process->priv->check_timeout);
			process->priv->check_timeout = 0;
		}

#if 0
		g_spawn_close_pid(process->priv->command_pid);
#endif
		process->priv->command_pid = 0;
		fr_channel_data_close_source (&process->out);
#if 0
		fr_channel_data_close_source (&process->err);
#endif

		process->priv->running = FALSE;

		if (emit_signal)
			g_signal_emit (G_OBJECT (process),
				       fr_process_signals[DONE],
				       0);
	}
}


void
fr_process_stop (FrProcess *process)
{
	fr_process_stop_priv (process, TRUE);
}
