#include <QtGui/QMessageBox>
#include <QTimer>

#include "mainserverwindow.h"
#include "ui_mainserverwindow.h"

#include "nntpgrab_utils.h"
#include "stdio.h"

#pragma GCC diagnostic ignored "-Wwrite-strings"
#include "icon_64x64.xpm"
#pragma GCC diagnostic warning "-Wwrite-strings"

void MainServerWindow::onPluginEvent(QString plugin_name, QString event_name, QList<QString> values)
{
    if (plugin_name == "JSON-RPC" && event_name == "num_active_connections_changed") {
        this->ui->lblNumActiveConnections->setText(values[0]);
    }
}

void MainServerWindow::onSchedularStateChanged(QNNTPGrab::SchedularState new_state, QString reason)
{
   switch (new_state) {
    case SCHEDULAR_STATE_RUNNING:
        this->ui->lblSchedularState->setText(QObject::tr("Running"));
        this->ui->btnPauseSchedular->setText(QObject::tr("Pause the schedular"));
        this->ui->btnPauseSchedular->setDisabled(false);
        break;

    case SCHEDULAR_STATE_STOPPING:
        this->ui->lblSchedularState->setText(QObject::tr("Stopping"));
        this->ui->btnPauseSchedular->setDisabled(true);
        break;

    case SCHEDULAR_STATE_STOPPED:
    default:
        this->ui->lblSchedularState->setText(QObject::tr("Stopped"));
        this->ui->btnPauseSchedular->setText(QObject::tr("Start the schedular"));
        this->ui->btnPauseSchedular->setDisabled(false);
        break;
    }

    if (!reason.isEmpty()) {
        QMessageBox b;
        b.setText(reason);

        b.show();
        b.exec();
    }
}

void MainServerWindow::onLogMessage(QString component, QNNTPGrab::LogLevel log_level, QString message)
{
    QTextEdit *txt = this->ui->txtServerLog;
    QString str;
    QString log_level_str;

    switch (log_level) {
    case NG_LOG_LEVEL_INFO:
        log_level_str = "info";
        break;

    case NG_LOG_LEVEL_WARNING:
        log_level_str = "warning";
        break;

    case NG_LOG_LEVEL_ERROR:
        log_level_str = "error";
        break;

    case NG_LOG_LEVEL_FATAL:
        log_level_str = "fatal";
        break;

    case NG_LOG_LEVEL_DEBUG:
        log_level_str = "debug";
        break;

    case NG_LOG_LEVEL_ALL:
    default:
        /* Shouldn't happen */
        log_level_str = "unknown";
        break;
    }

    /* TODO: Use colors to indicate the different log levels */
    txt->append(component + " " + log_level_str + " " + message);

    if (log_level == QNNTPGrab::LOG_LEVEL_WARNING) {
        QMessageBox b;

        b.setText(component + " " + message);
        b.show();
        b.exec();
    }
}

MainServerWindow::MainServerWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainServerWindow)
{
    ui->setupUi(this);

    /* Initialize the NNTPGrab library as soon as we've entered the main loop */
    QTimer::singleShot(0, this, SLOT(initNNTPGrab()));

    /* Give the window a nice icon */
    QPixmap pixmap(icon_64x64_xpm);
    QIcon icon(pixmap);

    this->setWindowIcon(icon);

    /* Add a tray icon if possible */
    /* But not on Mac OS X as tray icon's aren't commonly used there */
#ifndef Q_OS_DARWIN
    if (QSystemTrayIcon::isSystemTrayAvailable()) {
        trayIcon.setIcon(icon);
        trayIcon.setToolTip("NNTPGrab Server");
        trayIcon.show();
        connect(&trayIcon,
                SIGNAL(activated(QSystemTrayIcon::ActivationReason)),
                SLOT(trayIconActivated(QSystemTrayIcon::ActivationReason)));
    }
#endif
}

MainServerWindow::~MainServerWindow()
{
    delete ui;
}

#ifndef Q_OS_DARWIN
void MainServerWindow::trayIconActivated(QSystemTrayIcon::ActivationReason reason)
{
    if (reason == QSystemTrayIcon::Trigger) {
        if (this->isVisible() && this->isActiveWindow()) {
            /* If we're already being shown AND on the foreground, hide */
            this->hide();
        } else if (this->isVisible() && !this->isActiveWindow()) {
            /* If we're already being shown AND NOT on the foreground, bring to the foreground */
            this->raise();
            this->activateWindow();
        } else if (!this->isVisible()) {
            /* If we're NOT being shown, show */
            this->show();
            this->raise();
            this->activateWindow();
        }
    }
}
#endif

void MainServerWindow::changeEvent(QEvent *e)
{
    QMainWindow::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

#ifdef Q_WS_WIN
#include <windows.h>
bool MainServerWindow::winEvent(MSG* msg, long* result)
{
    if (msg->message == WM_SYSCOMMAND && msg->wParam == SC_MINIMIZE) {
        hide(); // minimized toolwindows look bad on Windows, so let's just hide it instead
                // plus we cannot do this in changeEvent(), because it's called too late
        *result = 0;
        return true;    // don't let Qt process this event
    }
    return false;
}
#elif defined(Q_WS_X11)
void MainServerWindow::hideEvent(QHideEvent *e)
{
    QMainWindow::hideEvent(e);

    /* Hide the program from the task bar when the tray icon is visible */
    if (QSystemTrayIcon::isSystemTrayAvailable()) {
        this->hide();
    }
}
#endif

void MainServerWindow::initNNTPGrab()
{
    char *err = NULL;
    char *warnings = NULL;

    if (!nntpgrab.init(&err, &warnings)) {
        QMessageBox b;

        b.setText(QObject::tr("Unable to initialize NNTPGrab: ") + err);
        ng_free(err);

        b.show();
        b.exec();

        close();

        return;
    }

    if (warnings) {
        QMessageBox b;

        b.setText(warnings);
        ng_free(warnings);

        b.show();
        b.exec();
    }

    this->nntpgrab.startSchedular();

    connect(&this->nntpgrab,
            SIGNAL(pluginEvent(QString, QString, QList<QString>)),
            SLOT(onPluginEvent(QString, QString, QList<QString>)));

    connect(&this->nntpgrab,
            SIGNAL(schedularStateChanged(QNNTPGrab::SchedularState, QString)),
            SLOT(onSchedularStateChanged(QNNTPGrab::SchedularState, QString)));

    connect(&this->nntpgrab,
            SIGNAL(logMessage(QString, QNNTPGrab::LogLevel, QString)),
            SLOT(onLogMessage(QString, QNNTPGrab::LogLevel, QString)));

    /* Start the embedded JSON-RPC/Webserver so other clients can connect to us */
    char *errmsg = NULL;
    ConfigOpts opts = this->nntpgrab.configGetOpts();
    if (!this->nntpgrab.startEmbeddedServer(opts.webserver_port, &errmsg)) {
        QMessageBox b;

        b.setText("Unable to start the embedded server: " + QString(errmsg));
        b.show();
        b.exec();

        ng_free(errmsg);

        close();
    }
}

void MainServerWindow::on_btnPauseSchedular_clicked()
{
    switch (this->nntpgrab.getSchedularState()) {
    case SCHEDULAR_STATE_RUNNING:
        this->nntpgrab.stopSchedular(FALSE);
        break;

    case SCHEDULAR_STATE_STOPPED:
        this->nntpgrab.startSchedular();
        break;

    case SCHEDULAR_STATE_STOPPING:
        /* Do nothing */
        break;
    }
}

void MainServerWindow::on_btnExit_clicked()
{
    close();
}

